
/*
 * libgnet - game net library
 * Copyright (C) 1997 Indrek Mandre 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <string.h>
#include <errno.h>

#include "gnet.h"
#include "gnet-local.h"

void GN_init_clients ( GN_server *server )
{

 server->clients = malloc ( sizeof ( GN_client ) );
 server->clients->next = NULL;
 server->clients->prev = NULL;

}

GN_client * GN_add_client ( GN_server *server )
{

 server->clients->next = malloc ( sizeof ( GN_client ) );
 server->clients->next->prev = server->clients;
 server->clients->next->next = NULL;

 server->clients->server = server;
 server->clients = server->clients->next;

 return server->clients->prev;

}

GN_client * GN_rem_client ( GN_client *client )
{

 if ( client->udpaddress != NULL ) free ( client->udpaddress );
 close ( client->tcpsocket );
 if ( client->udpsocket != -1 ) close ( client->udpsocket );

 if ( client->prev == NULL ) {
 	client = client->next;
 	free ( client->prev );
	client->prev = NULL;
	return client;
 } else {
 	GN_client *tmp1;
 	GN_client *tmp2;
	tmp1 = client->next;
	tmp2 = client->prev;
	tmp1->prev = tmp2;
	tmp2->next = tmp1;
	free ( client );
	return tmp1;
 }

}

GN_client * GN_accept_client ( GN_server *server )
{

 if ( server->tcpsocktype == NET_UNIX ) {
 	struct GN_connection con;
 	struct sockaddr_un un;
 	GN_client *cli;
 	int size = sizeof ( struct sockaddr_un );
 	int s;

	memset ( &un, 0, sizeof ( struct sockaddr_un ) );
	s = accept ( server->tcpsocket, (struct sockaddr *)&un, &size ); 

 	if ( s < 0 ) {
 		sprintf ( GN_error, "Error accepting socket, errno=%s\n", strerror ( errno ) );
 		return NULL;
 	}

 	cli = GN_add_client ( server );
 	cli->tcpsocket = s;
 	cli->udpsocket = -1;
 	strcpy ( cli->fromadd, "127.0.0.1" );
 	cli->fromport = 0;
 	cli->flags = FLAG_WAIT_ANSWER;

 	con.version = GNET_VERSION;
 	con.udpport = -1;

 	send ( cli->tcpsocket, &con, sizeof ( struct GN_connection ), 0 );

 	return cli;

 } else {
 	struct GN_connection con;
 	struct sockaddr_in in;
 	struct sockaddr_in udp_in;
 	GN_client *cli;
 	int udp = -1;
 	int size;
 	int len;
 	int s;

 	if ( server->isudp ) {
 		if ( ( udp = socket ( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
 			sprintf ( GN_error, "Couldn't create udp socket, errno=%s\n", strerror ( errno ) );
 			return NULL;
 		}

 		memset ((char *)&udp_in, 0, sizeof(struct sockaddr_in));
 		udp_in.sin_family = AF_INET;
 		udp_in.sin_addr.s_addr = INADDR_ANY;
 		udp_in.sin_port = 0;

 		len = sizeof ( struct sockaddr_in );

		if ( bind ( udp, (struct sockaddr *)&udp_in, sizeof ( struct sockaddr_in ) ) == -1 ) {
 			sprintf ( GN_error, "Couldn't bind udp socket, errno=%s\n", strerror ( errno ) );
 			return NULL;
		}

		getsockname ( udp, (struct sockaddr *) &udp_in, &len );
 	}

	memset ( &in, 0, sizeof ( struct sockaddr_in ) );

	size = sizeof ( struct sockaddr_in );

 	s = accept ( server->tcpsocket, (struct sockaddr *)&in, &size );

 	if ( s < 0 ) {
 		sprintf ( GN_error, "Error accepting socket, errno=%s\n", strerror ( errno ) );
 		if ( server->isudp ) close ( udp );
 		return NULL;
 	}

 	cli = GN_add_client ( server );
 	cli->tcpsocket = s;
 	cli->udpsocket = udp;
 	cli->udpaddress = NULL;

 	if ( server->isudp ) {
 		cli->udpaddress = malloc ( sizeof ( struct sockaddr_in ) );
 		memset ( cli->udpaddress, 0, sizeof ( struct sockaddr_in ) );
 		((struct sockaddr_in *)cli->udpaddress)->sin_family = AF_INET;
 		((struct sockaddr_in *)cli->udpaddress)->sin_port = 0;
 		((struct sockaddr_in *)cli->udpaddress)->sin_addr.s_addr = in.sin_addr.s_addr;
 	}

 	strcpy ( cli->fromadd, inet_ntoa ( in.sin_addr ) );
 	cli->fromport = htons ( in.sin_port );

 	cli->flags = FLAG_WAIT_ANSWER;

 	con.version = GNET_VERSION;

 	if ( server->isudp ) con.udpport = htons ( udp_in.sin_port );
 		else con.udpport = -1;

 	send ( cli->tcpsocket, &con, sizeof ( struct GN_connection ), 0 );

 	return cli;
 }

}

